/*
 * ESP32 Spiral Lamp
 *
 * Made from a lamp purchased from Home Depot
 *
 * This code controls a modern spiral lamp
 * which lights up whenever motion is detected.
 *
 * The lamp consists of two strands of WS2812B LEDs. Strand 1
 * is the shorter strand with 61 LEDs. Strand 2 is the
 * longer strand with 70 LEDs. FastLED is configured to
 * treat both strands as one long one.
 *
 * The ESP32 controller sleeps until motion is
 * detected, then lights up for LIGHT_DISPLAY_SECS
 * then goes back to sleep awaiting further motion.
 *
 * If the mode control pushbutton is held while power is
 * applied or when the light wakes up from sleeping,
 * the mode control state will be toggled:
 * off to on or on to off. If mode control is OFF,
 * the lamp will only light up when motion is
 * detected. If mode control is ON, the lamp will 
 * also light up every SLEEP_TIME_IN_MINS, without motion
 * being required.
 *
 * The motion control state is held in flash memory (EEPROM)
 * so it survives power cycling.
 *
 * Hardware:
 *  WEMOS D1 Mini ESP32 Module
 *  131 WS2812B NeoPixels
 *  HC-SR04 PIR motion sensor module
 *  Pushbutton (mode control pushbutton)
 *  A IN4001 diode to separate USB power from power
 *  supply power
 *  A 5V 5 amp (or more) power supply
 *
 * Compile using the WEMOS D1 Mini ESP32 board type 
 * Uses the FastLED library
 *
 * Concept, Design and Implementation by: Craig A. Lindley
 * Last Update: 02/10/2025
 */

// Include FastLED library
#include <FastLED.h>

// Include library to read and write from flash memory
#include <EEPROM.h>

// Include library for adding entropy to random number generator
#include <bootloader_random.h>

// Pin driving the shorter WS2812B strand 1
#define LED_STRAND_1_PIN 19

// Count of WS2812B LEDs on strand 1
#define LED_STRAND_1_COUNT 61

// Pin driving the longer WS2812B strand 2
#define LED_STRAND_2_PIN 18

// Count of WS2812B LEDs on strand 2
#define LED_STRAND_2_COUNT 70

// Total count of LEDs
#define TOTAL_LED_COUNT (LED_STRAND_1_COUNT + LED_STRAND_2_COUNT)

// PIR sensor is connected to GPIO pin 25
// Only RTC GPIO pins can be used for waking
#define PIR_PIN GPIO_NUM_25

// Mode control toggle pushbutton pin
#define MC_PIN 22

// Mode control toggle byte is the 0th or first byte of EEPROM memory
#define TOGGLE_BYTE_ADDR 0

// Conversion factor from seconds to micro seconds
#define SECS_TO_USECS 1000000

// Timer wake up timing
#define SLEEP_TIME_IN_MINS 4
#define SLEEP_TIME_IN_SECS (SLEEP_TIME_IN_MINS * 60)

// Light display timing
#define LIGHT_DISPLAY_SECS 90
#define LIGHT_DISPLAY_MSECS (LIGHT_DISPLAY_SECS * 1000)

// A delay to keep the frame rate modest
#define FRAMES_PER_SECOND 60

/**************************************************************************/
/***         Forward references to all of the display patterns          ***/
/**************************************************************************/

void pattern0();
void pattern1();
void pattern2();
void pattern3();
void pattern4();
void pattern5();
void pattern6();

/**************************************************************************/
/***                     Misc Constants and Variables                   ***/
/**************************************************************************/

// Buffer for all of the LED data
CRGB leds[TOTAL_LED_COUNT];

// Global hue value used by some of the patterns
uint8_t gHue = 0;

// Controls direction of lighting in some patterns
bool reverseDirection = false;

/**************************************************************************/
/***                          Support Functions                        ***/
/**************************************************************************/

// A function pointer for executing display patterns from an array
typedef void (*pt2Function)();

// Array of pointers to pattern display functions
pt2Function patternFunctions[] = {

  pattern0, pattern1, pattern2, pattern3,
  pattern4, pattern5, pattern6
};

// Determine the number of display patterns from the entries in the array
#define NUMBER_OF_PATTERNS (sizeof(patternFunctions) / sizeof(pt2Function))

/**************************************************************************/
/***                           Pattern Functions                        ***/
/**************************************************************************/

// NOTE: pattern functions are called for each frame of  pattern

void pattern0() {

  // FastLED's built-in rainbow generator
  fill_rainbow(leds, TOTAL_LED_COUNT, gHue, 12);
}

void pattern1() {

  // FastLED's built-in rainbow generator plus some random sparkly glitter
  fill_rainbow(leds, TOTAL_LED_COUNT, gHue, 12);

  // Add some glitter
  if (random8() < 80) {
    leds[random16(TOTAL_LED_COUNT)] += CRGB::White;
  }
}

void pattern2() {

  // Confetti
  // Random colored speckles that blink in and fade smoothly
  fadeToBlackBy(leds, TOTAL_LED_COUNT, 10);
  int pos = random16(TOTAL_LED_COUNT);
  leds[pos] += CHSV(gHue + random8(64), 200, 255);
}

void pattern3() {

  // Sinelon
  // A colored dot sweeping back and forth, with fading trails
  fadeToBlackBy(leds, TOTAL_LED_COUNT, 20);
  int pos = beatsin16(13, 0, TOTAL_LED_COUNT - 1);
  leds[pos] += CHSV(gHue, 255, 192);
}

void pattern4() {

  // BPM
  // Colored stripes pulsing at a defined Beats-Per-Minute (BPM)
  uint8_t BeatsPerMinute = 62;
  CRGBPalette16 palette = PartyColors_p;
  uint8_t beat = beatsin8(BeatsPerMinute, 64, 255);
  for (int i = 0; i < TOTAL_LED_COUNT; i++) {
    leds[i] = ColorFromPalette(palette, gHue + (i * 2), beat - gHue + (i * 10));
  }
}

void pattern5() {

  // Juggle
  // Eight colored dots, weaving in and out of sync with each other
  fadeToBlackBy(leds, TOTAL_LED_COUNT, 20);
  uint8_t dothue = 0;
  for (int i = 0; i < 8; i++) {
    leds[beatsin16(i + 7, 0, TOTAL_LED_COUNT - 1)] |= CHSV(dothue, 200, 255);
    dothue += 32;
  }
}

// Fire2012 by Mark Kriegsman, July 2012
//
// This basic one-dimensional 'fire' simulation works roughly as follows:
// There's a underlying array of 'heat' cells, that model the temperature
// at each point along the line.  Every cycle through the simulation,
// four steps are performed:
//  1) All cells cool down a little bit, losing heat to the air
//  2) The heat from each cell drifts 'up' and diffuses a little
//  3) Sometimes randomly new 'sparks' of heat are added at the bottom
//  4) The heat from each cell is rendered as a color into the leds array
//     The heat-to-color mapping uses a black-body radiation approximation.
//
// Temperature is in arbitrary units from 0 (cold black) to 255 (white hot).
//
// This simulation scales it self a bit depending on TOTAL_LED_COUNT; it should look
// "OK" on anywhere from 20 to 100 LEDs without too much tweaking.
//
// I recommend running this simulation at anywhere from 30-100 frames per second,
// meaning an interframe delay of about 10-35 milliseconds.
//
// Looks best on a high-density LED setup (60+ pixels/meter).
//
// There are two main parameters you can play with to control the look and
// feel of your fire: COOLING (used in step 1 above), and SPARKING (used
// in step 3 above).
//
// COOLING: How much does the air cool as it rises?
// Less cooling = taller flames.  More cooling = shorter flames.
// Default 50, suggested range 20-100
#define COOLING 75

// SPARKING: What chance (out of 255) is there that a new spark will be lit?
// Higher chance = more roaring fire.  Lower chance = more flickery fire.
// Default 120, suggested range 50-200.
#define SPARKING 60

void pattern6() {

  // Array of temperature readings at each simulation cell
  static uint8_t heat[TOTAL_LED_COUNT];

  // Step 1.  Cool down every cell a little
  for (int i = 0; i < TOTAL_LED_COUNT; i++) {
    heat[i] = qsub8(heat[i], random8(0, ((COOLING * 10) / TOTAL_LED_COUNT) + 2));
  }

  // Step 2.  Heat from each cell drifts 'up' and diffuses a little
  for (int k = TOTAL_LED_COUNT - 1; k >= 2; k--) {
    heat[k] = (heat[k - 1] + heat[k - 2] + heat[k - 2]) / 3;
  }

  // Step 3.  Randomly ignite new 'sparks' of heat near the bottom
  if (random8() < SPARKING) {
    int y = random8(7);
    heat[y] = qadd8(heat[y], random8(160, 255));
  }

  // Step 4.  Map from heat cells to LED colors
  for (int j = 0; j < TOTAL_LED_COUNT; j++) {
    CRGB color = HeatColor(heat[j]);
    int pixelnumber;
    if (reverseDirection) {
      pixelnumber = (TOTAL_LED_COUNT - 1) - j;
    } else {
      pixelnumber = j;
    }
    leds[pixelnumber] = color;
  }
}

/**************************************************************************/
/***                            Program Setup                           ***/
/**************************************************************************/

void setup() {

  Serial.begin(115200);
  delay(1000);
  Serial.println("\nWaking Up");

  // Instantiate the WS2812B driver for strand 1
  FastLED.addLeds<WS2812B, LED_STRAND_1_PIN, GRB>(leds, 0, LED_STRAND_1_COUNT);

  // Instantiate the WS2812B driver for strand 2
  FastLED.addLeds<WS2812B, LED_STRAND_2_PIN, GRB>(leds, LED_STRAND_1_COUNT, LED_STRAND_2_COUNT);

  // Set overall brightness of strands
  FastLED.setBrightness(64);

  // Limit current draw
  FastLED.setMaxPowerInVoltsAndMilliamps(5, 5000);

  // Increase randomness
  bootloader_random_enable();

  // Initialize EEPROM library, allocate 1 byte for use
  EEPROM.begin(1);

  // Configure Mode control toggle pin
  pinMode(MC_PIN, INPUT_PULLUP);
  delay(50);

  // Is the mode control pushbutton active ?
  // If so toggle saved state
  if (digitalRead(MC_PIN) == LOW) {
    Serial.println("MC active");

    // Switch is active so toggle saved state
    if (EEPROM.read(TOGGLE_BYTE_ADDR) == 0) {
      // Write a value of 1
      EEPROM.write(TOGGLE_BYTE_ADDR, 1);
    } else {
      // Write a value of 0
      EEPROM.write(TOGGLE_BYTE_ADDR, 0);
    }
    // Commit the new value
    EEPROM.commit();
  }

  // Configure GPIO input pin to read sensor
  pinMode(PIR_PIN, INPUT);

  // Configure ESP32 to wake when PIR sensor detects motion
  esp_sleep_enable_ext0_wakeup(PIR_PIN, HIGH);

  // Randomly pick a direction
  // Not many patterns use this
  reverseDirection = (random(2) == 0);

  // Pick random pattern index
  int pi = random(NUMBER_OF_PATTERNS);
  // pi = 6;  // For testing

  // Determine time in the future for pattern display to end
  unsigned long futureTime = millis() + LIGHT_DISPLAY_MSECS;

  // Run until time display time is up
  while (futureTime > millis()) {

    // Slowly cycle the hue through the rainbow
    EVERY_N_MILLISECONDS(20) {
      gHue++;
    }

    // Execute the selected pattern
    (*patternFunctions[pi])();

    // Display the pattern created above
    FastLED.show();

    // Insert a delay to keep the frame rate modest
    FastLED.delay(1000 / FRAMES_PER_SECOND);
  }

  // Turn all the LEDs off
  FastLED.clear();
  FastLED.show();

  // Possibly configure ESP32 to wake up from timer as well
  if (EEPROM.read(TOGGLE_BYTE_ADDR) == 1) {
    Serial.println("Auto");
    esp_sleep_enable_timer_wakeup(SLEEP_TIME_IN_SECS * SECS_TO_USECS);
  }

  Serial.println("Going to Sleep");

  // Put the ESP32 to sleep
  esp_deep_sleep_start();
}

/**************************************************************************/
/***                             Program Loop                           ***/
/**************************************************************************/

// Will never get here
void loop() {
}
